#!/usr/bin/env python
# -*- coding: utf-8 -*-
import random
import base64
import unittest
import pkg_resources

import asyncio
import aiounittest

from .test_data import XmrTestData
from .. import xmrserialize as x
from .. import xmrtypes as xmr


__author__ = 'dusanklinec'


class XmrTypesBaseTest(aiounittest.AsyncTestCase):
    """Simple tests"""

    def __init__(self, *args, **kwargs):
        super(XmrTypesBaseTest, self).__init__(*args, **kwargs)
        self.test_data = XmrTestData()

    def setUp(self):
            self.test_data.reset()

    async def test_simple_msg(self):
        """
        TxinGen
        :return:
        """
        msg = xmr.TxinGen(height=42)

        writer = x.MemoryReaderWriter()
        ar1 = x.Archive(writer, True)
        await ar1.message(msg)

        msg2 = xmr.TxinGen()
        ar2 = x.Archive(x.MemoryReaderWriter(writer.buffer), False)
        await ar2.message(msg2)

        self.assertEqual(msg.height, msg2.height)
        self.assertEqual(msg, msg2)

    async def test_boro_sig(self):
        """
        BoroSig
        :return:
        """
        msg = self.test_data.gen_borosig()

        writer = x.MemoryReaderWriter()
        ar1 = x.Archive(writer, True)
        await ar1.message(msg)

        msg2 = xmr.BoroSig()
        ar2 = x.Archive(x.MemoryReaderWriter(writer.buffer), False)
        await ar2.message(msg2)

        self.assertEqual(msg, msg2)




if __name__ == "__main__":
    unittest.main()  # pragma: no cover






#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Monero Boost codec, portable binary archive
'''

import binascii

from . import xmrserialize as x


_UVARINT_BUFFER = bytearray(1)


async def load_uvarint(reader):
    """
    Monero portable_binary_archive boost integer serialization
    :param reader:
    :return:
    """
    buffer = _UVARINT_BUFFER
    await reader.areadinto(buffer)
    size = buffer[0]
    if size == 0:
        return 0

    negative = size < 0
    size = -size if negative else size
    result = 0
    shift = 0

    if size > 8:
        raise ValueError('Varint size too big')

    # TODO: endianity, rev bytes if needed
    for _ in range(size):
        await reader.areadinto(buffer)
        result += buffer[0] << shift
        shift += 8
    return result if not negative else -result


async def dump_uvarint(writer, n):
    """
    Monero portable_binary_archive boost integer serialization
    :param writer:
    :param n:
    :return:
    """
    buffer = _UVARINT_BUFFER
    if n == 0:
        buffer[0] = 0
        return await writer.awrite(buffer)

    negative = n < 0
    ll = -n if negative else n

    size = 0
    while ll != 0:
        ll >>= 8
        size += 1

    buffer[0] = size
    await writer.awrite(buffer)

    ll = -n if negative else n

    # TODO: endianity, rev bytes if needed
    for _ in range(size):
        buffer[0] = n & 0xff
        await writer.awrite(buffer)
        ll >>= 8


class TypeWrapper(object):
    """
    Boost serialization type wrapper - versioning.
    Primitive types not versioned.
    """
    ELEMENTARY_RES = 0, 0

    def __init__(self, tp, params=None):
        self.tp = tp
        self.params = TypeWrapper.wrap_params(params)

    @staticmethod
    def is_elementary_type(elem_type):
        """
        Returns True if the type is elementary - not versioned
        :param elem_type:
        :return:
        """
        if not x.is_type(elem_type, x.XmrType):
            return False
        if x.is_type(elem_type, (x.UVarintType, x.IntType, x.UnicodeType)):
            return True
        return False

    @staticmethod
    def wrap_params(params):
        if params is None:
            return None
        if isinstance(params, (tuple, list)) and len(params) == 0:
            return None
        if not isinstance(params, (tuple, list)):
            params = (params, )
        return params

    def is_elementary(self):
        return TypeWrapper.is_elementary_type(self.tp)

    def is_versioned(self):
        if TypeWrapper.is_elementary_type(self.tp):
            return False
        if hasattr(self.tp, 'boost_versioned'):
            return self.tp.boost_versioned()
        if hasattr(self.tp, 'VERSIONED'):
            return self.VERSIONED
        else:
            return True

    def get_current_version(self):
        if hasattr(self.tp, 'boost_version'):
            return self.tp.boost_version()
        if hasattr(self.tp, 'VERSION'):
            return self.VERSION
        else:
            return 0

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False
        return self.tp == other.tp and self.params == other.params

    def __hash__(self):
        return hash(self.tp) ^ hash(self.params)

    def __repr__(self):
        return 'Type<%r:%r>' % (self.tp, self.params)


class VersionDatabase(object):
    """
    Boost version database - singletons
    """
    def __init__(self):
        self.db = {}  # type: dict[type -> tuple[int, int]]

    def is_versioned(self, twrap):
        """
        Returns true if type wrapper is versioned
        :param twrap:
        :return:
        """
        return twrap in self.db

    def get_version(self, twrap):
        return self.db[twrap]

    def set_version(self, twrap, track, version):
        if self.is_versioned(twrap):
            return False
        self.db[twrap] = (track, version)


class Archive(x.Archive):

    def __init__(self, iobj, writing=True, **kwargs):
        super().__init__(iobj, writing, **kwargs)
        self.version_db = VersionDatabase()

    def type_in_db(self, tp, params):
        """
        Determines if type is in the database
        :param tp:
        :return:
        """
        tw = TypeWrapper(tp, params)
        return self.version_db.is_versioned(tw)

    async def get_version(self, tp, params):
        """
        Loads version from the stream / version database
        # TODO: instance vs. tp.

        :param tp:
        :param params:
        :return:
        """
        tw = TypeWrapper(tp, params)
        if not tw.is_versioned():
            return TypeWrapper.ELEMENTARY_RES

        # If not in the DB, load from archive at current position
        if not self.version_db.is_versioned(tw):
            tr = await load_uvarint(self.iobj)
            if tr != 0:
                raise ValueError('Unsupported tracking for %s, tr: %s' % (tw, tr))

            ver = await load_uvarint(self.iobj)

            self.version_db.set_version(tw, tr, ver)

        return self.version_db.get_version(tw)[1]

    async def set_version(self, tp, params, version=None):
        """
        Stores version to the stream if not stored yet

        :param tp:
        :param params:
        :param version:
        :return:
        """
        tw = TypeWrapper(tp, params)
        if not tw.is_versioned():
            return TypeWrapper.ELEMENTARY_RES

        # If not in the DB, store to the archive at the current position
        if not self.version_db.is_versioned(tw):
            if version is None:
                version = tw.get_current_version()
            await dump_uvarint(self.iobj, 0)
            await dump_uvarint(self.iobj, version)
            self.version_db.set_version(tw, 0, version)

        return self.version_db.get_version(tw)[1]

    async def version(self, tp, params):
        """
        Symmetric version management

        :param tp:
        :param params:
        :return:
        """
        if self.writing:
            return await self.set_version(tp, params, params)
        else:
            return await self.get_version(tp, params)

    async def uvarint(self, elem):
        """
        Uvarint
        :param elem:
        :return:
        """
        if self.writing:
            return await dump_uvarint(self.iobj, elem)
        else:
            return await load_uvarint(self.iobj)

    async def uint(self, elem, elem_type, params=None):
        """
        Integer types
        :param elem:
        :param elem_type:
        :param params:
        :return:
        """
        one_b_type = x.is_type(elem_type, (x.Int8, x.UInt8))
        if self.writing:
            return (await x.dump_uint(self.iobj, elem, 1)) if one_b_type else (await dump_uvarint(self.iobj, elem))
        else:
            return (await x.load_uint(self.iobj, 1)) if one_b_type else (await load_uvarint(self.iobj))

    async def unicode_type(self, elem):
        """
        Unicode type
        :param elem:
        :return:
        """
        if self.writing:
            await dump_uvarint(self.iobj, len(elem))
            await self.iobj.awrite(bytes(elem, 'utf8'))

        else:
            ivalue = await load_uvarint(self.iobj)
            if ivalue == 0:
                return ''

            fvalue = bytearray(ivalue)
            await self.iobj.areadinto(fvalue)
            return str(fvalue, 'utf8')

    async def blob(self, elem=None, elem_type=None, params=None):
        """
        Loads/dumps blob
        :return:
        """
        elem_type = elem_type if elem_type else elem.__class__
        version = await self.version(elem_type, params)

        if hasattr(elem_type, 'boost_serialize'):
            elem = elem_type() if elem is None else elem
            return await elem.boost_serialize(self, elem=elem, elem_type=elem_type, params=params, version=version)

        if self.writing:
            return await self.blob_dump(elem=elem, elem_type=elem_type, params=params)

        else:
            return await self.blob_load(elem_type=elem_type, params=params, elem=elem)

    async def blob_dump(self, elem, elem_type, params=None):
        """
        Dumps blob message to the writer.
        Supports both blob and raw value.

        :param writer:
        :param elem:
        :param elem_type:
        :param params:
        :return:
        """
        elem_is_blob = isinstance(elem, x.BlobType)
        elem_params = elem if elem_is_blob or elem_type is None else elem_type
        data = getattr(elem, x.BlobType.DATA_ATTR) if elem_is_blob else elem

        if len(data) != elem_params.SIZE:
            raise ValueError('Fixed size blob has not defined size: %s' % elem_params.SIZE)

        await dump_uvarint(self.iobj, len(elem))
        await self.iobj.awrite(data)

    async def blob_load(self, elem_type, params=None, elem=None):
        """
        Loads blob from reader to the element. Returns the loaded blob.

        :param reader:
        :param elem_type:
        :param params:
        :param elem:
        :return:
        """
        ivalue = await load_uvarint(self.iobj)
        fvalue = bytearray(ivalue)
        await self.iobj.areadinto(fvalue)

        if elem is None:
            return fvalue  # array by default

        elif isinstance(elem, x.BlobType):
            setattr(elem, elem_type.DATA_ATTR, fvalue)
            return elem

        else:
            elem.extend(fvalue)

        return elem

    async def container(self, container=None, container_type=None, params=None):
        """
        Loads/dumps container
        :return:
        """
        # Container versioning is a bit tricky, primitive type containers are not versioned.
        elem_type = x.container_elem_type(container_type, params)
        raw_container = container_is_raw(container_type, params)
        elem_elementary = TypeWrapper.is_elementary_type(elem_type)
        is_versioned = not elem_elementary and not raw_container

        version = await self.version(container_type, params) if is_versioned else None
        if hasattr(container_type, 'boost_serialize'):
            container = container_type() if container is None else container
            return await container.boost_serialize(self, elem=container, elem_type=container_type, params=params, version=version)

        # Container entry version + container
        if self.writing:
            return await self.container_dump(container, container_type, params)
        else:
            return await self.container_load(container_type, params=params, container=container)

    async def container_size(self, container_len=None, container_type=None, params=None):
        """
        Container size
        :param container_len:
        :param container_type:
        :param params:
        :return:
        """
        if hasattr(container_type, 'boost_serialize'):
            raise ValueError('not supported')

        if self.writing:
            if not container_type or not container_type.FIX_SIZE:
                await dump_uvarint(self.iobj, container_len)
                if not container_is_raw(container_type, params):
                    await dump_uvarint(self.iobj, 0)  # element version

            elif container_len != container_type.SIZE:
                raise ValueError('Fixed size container has not defined size: %s' % container_type.SIZE)

        else:
            raise ValueError('Not supported')

    async def container_val(self, elem, container_type, params=None):
        """
        Single cont value
        :param elem:
        :param container_type:
        :param params:
        :param field_archiver:
        :return:
        """
        if hasattr(container_type, 'boost_serialize'):
            raise ValueError('not supported')

        if self.writing:
            elem_type = params[0] if params else None
            if elem_type is None:
                elem_type = container_type.ELEM_TYPE

            await self.dump_field(elem, elem_type, params[1:] if params else None)

        else:
            raise ValueError('Not supported')

    async def container_dump(self, container, container_type, params=None):
        """
        Dumps container of elements to the writer.

        :param writer:
        :param container:
        :param container_type:
        :param params:
        :param field_archiver:
        :return:
        """
        await self.container_size(len(container), container_type)
        if not container_is_raw(container_type, params):
            await dump_uvarint(self.iobj, 0)  # element version

        elem_type = params[0] if params else None
        if elem_type is None:
            elem_type = container_type.ELEM_TYPE
        for elem in container:
            await self._dump_field(elem, elem_type, params[1:] if params else None)

    async def container_load(self, container_type, params=None, container=None):
        """
        Loads container of elements from the reader. Supports the container ref.
        Returns loaded container.

        :param container_type:
        :param params:
        :param container:
        :param field_archiver:
        :return:
        """
        raw_container = container_is_raw(container_type, params)
        c_len = await load_uvarint(self.iobj)
        elem_ver = await load_uvarint(self.iobj) if not raw_container else 0

        # if container and c_len != len(container):
        #     raise ValueError('Size mismatch')

        elem_type = x.container_elem_type(container_type, params)
        res = container if container else []
        for i in range(c_len):
            fvalue = await self._load_field(elem_type,
                                            params[1:] if params else None,
                                            x.eref(res, i) if container else None)
            if not container:
                res.append(fvalue)
        return res

    async def tuple(self, elem=None, elem_type=None, params=None):
        """
        Loads/dumps tuple
        :return:
        """
        version = await self.version(elem_type, params)
        if hasattr(elem_type, 'boost_serialize'):
            container = elem_type() if elem is None else elem
            return await container.boost_serialize(self, elem=elem, elem_type=elem_type, params=params, version=version)

        if self.writing:
            return await self.dump_tuple(elem, elem_type, params)
        else:
            return await self.load_tuple(elem_type, params=params, elem=elem)

    async def dump_tuple(self, elem, elem_type, params=None):
        """
        Dumps tuple of elements to the writer.

        :param elem:
        :param elem_type:
        :param params:
        :return:
        """
        if len(elem) != len(elem_type.FIELDS):
            raise ValueError('Fixed size tuple has not defined size: %s' % len(elem_type.FIELDS))

        elem_fields = params[0] if params else None
        if elem_fields is None:
            elem_fields = elem_type.FIELDS
        for idx, elem in enumerate(elem):
            await self._dump_field(elem, elem_fields[idx], params[1:] if params else None)

    async def load_tuple(self, elem_type, params=None, elem=None):
        """
        Loads tuple of elements from the reader. Supports the tuple ref.
        Returns loaded tuple.

        :param elem_type:
        :param params:
        :param elem:
        :return:
        """
        elem_fields = params[0] if params else None
        if elem_fields is None:
            elem_fields = elem_type.FIELDS

        if elem and len(elem_fields) != len(elem):
            raise ValueError('Size mismatch')

        res = elem if elem else []
        for i in range(len(elem_fields)):
            fvalue = await self._load_field(elem_fields[i],
                                            params[1:] if params else None,
                                            x.eref(res, i) if elem else None)
            if not elem:
                res.append(fvalue)
        return res

    async def variant(self, elem=None, elem_type=None, params=None):
        """
        Loads/dumps variant type
        :param elem:
        :param elem_type:
        :param params:
        :return:
        """
        elem_type = elem_type if elem_type else elem.__class__
        version = await self.version(elem_type, params)

        if hasattr(elem_type, 'boost_serialize'):
            elem = elem_type() if elem is None else elem
            return await elem.boost_serialize(self, elem=elem, elem_type=elem_type, params=params, version=version)

        if self.writing:
            return await self.dump_variant(elem=elem,
                                           elem_type=elem_type if elem_type else elem.__class__, params=params)
        else:
            return await self.load_variant(elem_type=elem_type if elem_type else elem.__class__,
                                           params=params, elem=elem)

    async def dump_variant(self, elem, elem_type=None, params=None):
        """
        Dumps variant type to the writer.
        Supports both wrapped and raw variant.

        :param writer:
        :param elem:
        :param elem_type:
        :param params:
        :return:
        """
        if isinstance(elem, x.VariantType) or elem_type.WRAPS_VALUE:
            await dump_uvarint(self.iobj, elem.variant_elem_type.VARIANT_CODE)
            await self._dump_field(getattr(elem, elem.variant_elem), elem.variant_elem_type)

        else:
            fdef = elem_type.find_fdef(elem_type.FIELDS, elem)
            await dump_uvarint(self.iobj, fdef[1].VARIANT_CODE)
            await self._dump_field(elem, fdef[1])

    async def load_variant(self, elem_type, params=None, elem=None, wrapped=None):
        """
        Loads variant type from the reader.
        Supports both wrapped and raw variant.

        :param reader:
        :param elem_type:
        :param params:
        :param elem:
        :param wrapped:
        :return:
        """
        is_wrapped = (isinstance(elem, x.VariantType) or elem_type.WRAPS_VALUE) if wrapped is None else wrapped
        if is_wrapped:
            elem = elem_type() if elem is None else elem

        tag = await load_uvarint(self.iobj)
        for field in elem_type.FIELDS:
            ftype = field[1]
            vcode = ftype.BOOST_VARIANT_CODE if hasattr(ftype, 'BOOST_VARIANT_CODE') else ftype.VARIANT_CODE
            if vcode == tag:
                fvalue = await self._load_field(ftype, field[2:], elem if not is_wrapped else None)
                if is_wrapped:
                    elem.set_variant(field[0], fvalue)
                return elem if is_wrapped else fvalue
        raise ValueError('Unknown tag: %s' % tag)

    async def root(self):
        """
        Root level init
        :return:
        """
        if self.writing:
            await self.iobj.awrite(binascii.unhexlify(b'011673657269616c697a6174696f6e3a3a6172636869766500000'))

        else:
            hdr = bytearray(2)
            await self.iobj.areadinto(hdr)
            if hdr != bytearray(b'\x01\x16'):
                raise ValueError('Unsupported header')

            hdr = bytearray(22)
            await self.iobj.areadinto(hdr)
            if hdr != bytearray(b'serialization::archive'):
                raise ValueError('Unrecognized magic header')

            tra = await load_uvarint(self.iobj)
            ver = await load_uvarint(self.iobj)
            if tra != 0:
                raise ValueError('Tracking not supported')

    async def root_message(self, msg, msg_type=None):
        """
        Root-level message. First entry in the archive.
        Archive headers processing

        :return:
        """
        await self.root()
        await self.message(msg, msg_type)

    async def message(self, msg, msg_type=None, use_version=None):
        """
        Loads/dumps message
        :param msg:
        :param msg_type:
        :param use_version:
        :return:
        """
        elem_type = msg_type if msg_type is not None else msg.__class__
        version = await self.version(elem_type, None) if use_version is None else use_version

        if hasattr(elem_type, 'boost_serialize'):
            msg = elem_type() if msg is None else msg
            return await msg.boost_serialize(self, version=version)

        if self.writing:
            return await self.dump_message(msg, msg_type=msg_type)
        else:
            return await self.load_message(msg_type, msg=msg)

    async def message_field(self, msg, field, fvalue=None):
        """
        Dumps/Loads message field
        :param msg:
        :param field:
        :param fvalue: explicit value for dump
        :return:
        """
        if self.writing:
            fname, ftype, params = field[0], field[1], field[2:]
            fvalue = getattr(msg, fname, None) if fvalue is None else fvalue
            await self._dump_field(fvalue, ftype, params)

        else:
            fname, ftype, params = field[0], field[1], field[2:]
            await self._load_field(ftype, params, x.eref(msg, fname))

    async def message_fields(self, msg, fields):
        """
        Load/dump individual message fields
        :param msg:
        :param fields:
        :param field_archiver:
        :return:
        """
        for field in fields:
            await self.message_field(msg, field)
        return msg

    async def dump_message(self, msg, msg_type=None):
        """
        Dumps message to the writer.

        :param msg:
        :param msg_type:
        :return:
        """
        mtype = msg.__class__ if msg_type is None else msg_type
        fields = mtype.FIELDS
        for field in fields:
            await self.message_field(msg=msg, field=field)

    async def load_message(self, msg_type, msg=None):
        """
        Loads message if the given type from the reader.
        Supports reading directly to existing message.

        :param msg_type:
        :param msg:
        :return:
        """
        msg = msg_type() if msg is None else msg
        fields = msg_type.FIELDS if msg_type else msg.__class__.FIELDS
        for field in fields:
            await self.message_field(msg, field)

        return msg

    async def field(self, elem=None, elem_type=None, params=None):
        """
        Archive field
        :param elem:
        :param elem_type:
        :param params:
        :return:
        """
        elem_type = elem_type if elem_type else elem.__class__
        fvalue = None
        if issubclass(elem_type, x.UVarintType):
            fvalue = await self.uvarint(x.get_elem(elem))

        elif issubclass(elem_type, x.IntType):
            fvalue = await self.uint(elem=x.get_elem(elem), elem_type=elem_type, params=params)

        elif issubclass(elem_type, x.BlobType):
            fvalue = await self.blob(elem=x.get_elem(elem), elem_type=elem_type, params=params)

        elif issubclass(elem_type, x.UnicodeType):
            fvalue = await self.unicode_type(x.get_elem(elem))

        elif issubclass(elem_type, x.VariantType):
            fvalue = await self.variant(elem=x.get_elem(elem), elem_type=elem_type, params=params)

        elif issubclass(elem_type, x.ContainerType):  # container ~ simple list
            fvalue = await self.container(container=x.get_elem(elem), container_type=elem_type, params=params)

        elif issubclass(elem_type, x.TupleType):  # tuple ~ simple list
            fvalue = await self.tuple(elem=x.get_elem(elem), elem_type=elem_type, params=params)

        elif issubclass(elem_type, x.MessageType):
            fvalue = await self.message(x.get_elem(elem), msg_type=elem_type)

        else:
            raise TypeError

        return fvalue if self.writing else x.set_elem(elem, fvalue)

    async def dump_field(self, writer, elem, elem_type, params=None):
        assert self.iobj == writer
        return await self.field(elem=elem, elem_type=elem_type, params=params)

    async def load_field(self, reader, elem_type, params=None, elem=None):
        assert self.iobj == reader
        return await self.field(elem=elem, elem_type=elem_type, params=params)

    async def _dump_field(self, elem, elem_type, params=None):
        return await self.field(elem=elem, elem_type=elem_type, params=params)

    async def _load_field(self, elem_type, params=None, elem=None):
        return await self.field(elem=elem, elem_type=elem_type, params=params)


def container_is_raw(container_type, params):
    """
    Returns true if container is statically allocated array
    :param container_type:
    :param params:
    :return:
    """
    return container_type.BOOST_RAW_ARRAY if hasattr(container_type, 'BOOST_RAW_ARRAY') else False




#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Extremely minimal streaming codec for a subset of protobuf.  Supports uint32,
bytes, string, embedded message and repeated fields.

For de-sererializing (loading) protobuf types, object with `AsyncReader`
interface is required:

>>> class AsyncReader:
>>>     async def areadinto(self, buffer):
>>>         """
>>>         Reads `len(buffer)` bytes into `buffer`, or raises `EOFError`.
>>>         """

For serializing (dumping) protobuf types, object with `AsyncWriter` interface is
required:

>>> class AsyncWriter:
>>>     async def awrite(self, buffer):
>>>         """
>>>         Writes all bytes from `buffer`, or raises `EOFError`.
>>>         """
'''

const = lambda x: x


_UVARINT_BUFFER = bytearray(1)


async def load_uvarint(reader):
    buffer = _UVARINT_BUFFER
    result = 0
    shift = 0
    byte = 0x80
    while byte & 0x80:
        await reader.areadinto(buffer)
        byte = buffer[0]
        result += (byte & 0x7F) << shift
        shift += 7
    return result


async def dump_uvarint(writer, n):
    buffer = _UVARINT_BUFFER
    shifted = True
    while shifted:
        shifted = n >> 7
        buffer[0] = (n & 0x7F) | (0x80 if shifted else 0x00)
        await writer.awrite(buffer)
        n = shifted


class LimitedReader:

    def __init__(self, reader, limit):
        self.reader = reader
        self.limit = limit

    async def areadinto(self, buf):
        if self.limit < len(buf):
            raise EOFError
        else:
            nread = await self.reader.areadinto(buf)
            self.limit -= nread
            return nread


class CountingWriter:

    def __init__(self):
        self.size = 0

    async def awrite(self, buf):
        nwritten = len(buf)
        self.size += nwritten
        return nwritten


class AHashWriter:

    def __init__(self, hasher, sub_writer=None):
        self.hasher = hasher
        self.sub_writer = sub_writer

    async def awrite(self, buf):
        self.hasher.update(buf)
        if self.sub_writer:
            await self.sub_writer.awrite(buf)
        return len(buf)

    def get_digest(self, *args) -> bytes:
        return self.hasher.digest(*args)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
import random
import base64
import unittest
import json
import pkg_resources

import asyncio
import aiounittest

from .test_data import XmrTestData
from .. import xmrserialize as x
from .. import xmrtypes as xmr
from .. import xmrobj as xmro
from .. import xmrjson as xmrjs


__author__ = 'dusanklinec'


class XmrJsonTest(aiounittest.AsyncTestCase):
    """Simple tests"""

    def __init__(self, *args, **kwargs):
        super(XmrJsonTest, self).__init__(*args, **kwargs)
        self.test_data = XmrTestData()

    def setUp(self):
            self.test_data.reset()

    async def test_simple_msg(self):
        """
        TxinGen
        :return:
        """
        msg = xmr.TxinGen(height=42)
        msg_dict = await xmro.dump_message(None, msg)
        js = xmrjs.json_dumps(msg_dict, indent=2)
        self.assertTrue(len(js) > 0)

        popo = json.loads(js)
        msg2 = await xmro.load_message(popo, msg.__class__)
        self.assertIsNotNone(msg2)
        self.assertEqual(msg, msg2)

    async def test_tx_prefix(self):
        """
        TransactionPrefix
        :return:
        """
        msg = self.test_data.gen_transaction_prefix()
        msg_dict = await xmro.dump_message(None, msg)
        js = xmrjs.json_dumps(msg_dict, indent=2)
        self.assertTrue(len(js) > 0)

        popo = json.loads(js)
        msg2 = await xmro.load_message(popo, msg.__class__)
        self.assertIsNotNone(msg2)
        self.assertEqual(msg, msg2)

    async def test_boro_sig(self):
        """
        BoroSig
        :return:
        """
        msg = self.test_data.gen_borosig()
        msg_dict = await xmro.dump_message(None, msg)
        js = xmrjs.json_dumps(msg_dict, indent=2)
        self.assertTrue(len(js) > 0)

        popo = json.loads(js)
        msg2 = await xmro.load_message(popo, msg.__class__)
        self.assertIsNotNone(msg2)
        self.assertEqual(msg, msg2)


if __name__ == "__main__":
    unittest.main()  # pragma: no cover






#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Minimal streaming codec for a Monero binary serialization.
Used for a binary serialization in blockchain and for hash computation for signatures.

Equivalent of BEGIN_SERIALIZE_OBJECT(), /src/serialization/serialization.h

- The wire binary format does not use tags. Structure has to be read from the binary stream
with the scheme specified in order to parse the structure.

- Heavily uses variable integer serialization - similar to the UTF8 or LZ4 number encoding.

- Supports: blob, string, integer types - variable or fixed size, containers of elements,
            variant types, messages of elements

For de-serializing (loading) types, object with `AsyncReader`
interface is required:

>>> class AsyncReader:
>>>     async def areadinto(self, buffer):
>>>         """
>>>         Reads `len(buffer)` bytes into `buffer`, or raises `EOFError`.
>>>         """

For serializing (dumping) types, object with `AsyncWriter` interface is
required:

>>> class AsyncWriter:
>>>     async def awrite(self, buffer):
>>>         """
>>>         Writes all bytes from `buffer`, or raises `EOFError`.
>>>         """
'''

from .protobuf import const, load_uvarint, dump_uvarint


_UINT_BUFFER = bytearray(1)


async def load_uint(reader, width):
    """
    Constant-width integer serialization
    :param reader:
    :param width:
    :return:
    """
    buffer = _UINT_BUFFER
    result = 0
    shift = 0
    for _ in range(width):
        await reader.areadinto(buffer)
        result += buffer[0] << shift
        shift += 8
    return result


async def dump_uint(writer, n, width):
    """
    Constant-width integer serialization
    :param writer:
    :param n:
    :param width:
    :return:
    """
    buffer = _UINT_BUFFER
    for _ in range(width):
        buffer[0] = n & 0xff
        await writer.awrite(buffer)
        n >>= 8


def uvarint_size(n):
    """
    Returns size in bytes n would occupy serialized as varint
    :param n:
    :return:
    """
    bts = 0 if n != 0 else 1
    while n:
        n >>= 7
        bts += 1
    return bts


def load_uvarint_b(buffer):
    """
    Variable int deserialization, synchronous from buffer.
    :param buffer:
    :return:
    """
    result = 0
    idx = 0
    byte = 0x80
    while byte & 0x80:
        byte = buffer[idx]
        result += (byte & 0x7F) << (7 * idx)
        idx += 1
    return result


def dump_uvarint_b(n):
    """
    Serializes uvarint to the buffer
    :param n:
    :return:
    """
    buffer = bytearray(uvarint_size(n))
    return dump_uvarint_b_into(n, buffer, 0)


def dump_uvarint_b_into(n, buffer, offset=0):
    """
    Serializes n as variable size integer to the provided buffer.
    Buffer has to ha
    :param n:
    :param buffer:
    :param offset:
    :return:
    """
    shifted = True
    while shifted:
        shifted = n >> 7
        buffer[offset] = (n & 0x7F) | (0x80 if shifted else 0x00)
        offset += 1
        n = shifted
    return buffer


def load_uint_b(buffer, width):
    """
    Loads fixed size integer from the buffer
    :param buffer:
    :return:
    """
    result = 0
    for idx in range(width):
        result += buffer[idx] << (8 * idx)
    return result


def dump_uint_b(n, width):
    """
    Serializes fixed size integer to the buffer
    :param n:
    :param width:
    :return:
    """
    buffer = bytearray(width)
    return dump_uvarint_b_into(n, width, buffer, 0)


def dump_uint_b_into(n, width, buffer, offset=0):
    """
    Serializes fixed size integer to the buffer
    :param n:
    :param width:
    :return:
    """
    for idx in range(width):
        buffer[idx+offset] = n & 0xff
        n >>= 8
    return buffer


def eq_obj_slots(l, r):
    """
    Compares objects with __slots__ defined
    :param l:
    :param r:
    :return:
    """
    for f in l.__slots__:
        if getattr(l, f, None) != getattr(r, f, None):
            return False
    return True


def eq_obj_contents(l, r):
    """
    Compares object contents, supports slots
    :param l:
    :param r:
    :return:
    """
    if l.__class__ is not r.__class__:
        return False
    if hasattr(l, '__slots__'):
        return eq_obj_slots(l, r)
    else:
        return l.__dict__ == r.__dict__


def slot_obj_dict(o):
    """
    Builds dict for o with __slots__ defined
    :param o:
    :return:
    """
    d = {}
    for f in o.__slots__:
        d[f] = getattr(o, f, None)
    return d


def is_type(x, types, full=False):
    """
    Returns true if x is of type in types tuple
    :param x:
    :param types:
    :param full:
    :return:
    """
    types = types if isinstance(types, tuple) else (types, )
    ins = isinstance(x, types)
    sub = issubclass(x, types)
    res = ins or sub
    return res if not full else (res, ins)


class XmrType:
    VERSION = 0


class UVarintType(XmrType):
    pass


class IntType(XmrType):
    WIDTH = 0
    SIGNED = 0
    VARIABLE = 0

    def __repr__(self):
        return '%s:<w: %s, sig: %s, var: %s>' % (self.__class__, self.WIDTH, self.SIGNED, self.VARIABLE)


class BoolType(IntType):
    WIDTH = 1


class UInt8(IntType):
    WIDTH = 1


class Int8(IntType):
    SIGNED = 1
    WIDTH = 1


class UInt16(IntType):
    WIDTH = 2


class Int16(IntType):
    SIGNED = 1
    WIDTH = 2


class UInt32(IntType):
    WIDTH = 4


class Int32(IntType):
    SIGNED = 1
    WIDTH = 4


class UInt64(IntType):
    WIDTH = 8


class SizeT(UInt64):
    WIDTH = 8


class Int64(IntType):
    SIGNED = 1
    WIDTH = 8


class BlobType(XmrType):
    """
    Binary data

    Represented as bytearray() or a list of values in data structures.
    Not wrapped in the BlobType, the BlobType is only a scheme descriptor.
    Behaves in the same way as primitive types

    Supports also the wrapped version (__init__, DATA_ATTR, eq, repr...),
    """
    DATA_ATTR = 'data'
    FIX_SIZE = 0
    SIZE = 0

    def __init__(self, *args, **kwargs):
        if len(args) > 1:
            raise ValueError()
        if len(args) > 0:
            setattr(self, self.DATA_ATTR, args[0])
        if 'SIZE' in kwargs:
            self.SIZE = kwargs['SIZE']
        if 'FIX_SIZE' in kwargs:
            self.FIX_SIZE = kwargs['FIX_SIZE']

    def __eq__(self, rhs):
        return eq_obj_contents(self, rhs)

    def __repr__(self):
        dct = slot_obj_dict(self) if hasattr(self, '__slots__') else self.__dict__
        return '<%s: %s>' % (self.__class__.__name__, dct)


class UnicodeType(XmrType):
    pass


class VariantType(XmrType):
    """
    Union of types, variant tags needed. is only one of the types. List in typedef, enum.
    Wraps the variant type in order to unambiguously support variant of variants.
    Supports also unwrapped value using type system to distinguish variants - simplifies the construction.
    """
    WRAPS_VALUE = False
    FIELDS = []

    def __init__(self, *args, **kwargs):
        self.variant_elem = None
        self.variant_elem_type = None

        fname, fval = None, None
        if len(args) > 0:
            fname, fval = self.find_fdef(self.FIELDS, args[0])[0], args[0]
        if len(kwargs) > 0:
            key = list(kwargs.keys())[0]
            fname, fval = key, kwargs[key]
        if fname:
            self.set_variant(fname, fval)

    @staticmethod
    def find_fdef(fields, elem):
        for x in fields:
            if isinstance(elem, x[1]):
                return x
        raise ValueError('Unrecognized variant')

    def set_variant(self, fname, fvalue):
        self.variant_elem = fname
        self.variant_elem_type = fvalue.__class__
        setattr(self, fname, fvalue)

    def __eq__(self, rhs):
        return eq_obj_contents(self, rhs)

    def __repr__(self):
        dct = slot_obj_dict(self) if hasattr(self, '__slots__') else self.__dict__
        return '<%s: %s>' % (self.__class__.__name__, dct)


class ContainerType(XmrType):
    """
    Array of elements
    Represented as a real array in the data structures, not wrapped in the ContainerType.
    The Container type is used only as a schema descriptor for serialization.
    """
    FIX_SIZE = 0
    SIZE = 0
    ELEM_TYPE = None

    def __init__(self, *args, **kwargs):
        if 'SIZE' in kwargs:
            self.SIZE = kwargs['SIZE']
        if 'FIX_SIZE' in kwargs:
            self.FIX_SIZE = kwargs['FIX_SIZE']
        if 'ELEM_TYPE' in kwargs:
            self.ELEM_TYPE = kwargs['ELEM_TYPE']


class TupleType(XmrType):
    FIELDS = []  # simple types without file name

    def __init__(self, *args, **kwargs):
        if 'FIELDS' in kwargs:
            self.FIELDS = kwargs['FIELDS']


class MessageType(XmrType):
    FIELDS = []

    def __init__(self, **kwargs):
        for kw in kwargs:
            setattr(self, kw, kwargs[kw])

    def __eq__(self, rhs):
        return eq_obj_contents(self, rhs)

    def __repr__(self):
        dct = slot_obj_dict(self) if hasattr(self, '__slots__') else self.__dict__
        return '<%s: %s>' % (self.__class__.__name__, dct)

    def _field(self, fname=None, idx=None):
        fld = None
        if fname is not None:
            fld = [x for x in self.FIELDS if x[0] == fname][0]
        elif idx is not None:
            fld = self.FIELDS[idx]
        return fld

    async def _msg_field(self, ar, fname=None, idx=None):
        return await ar.message_field(self, self._field(fname=fname, idx=idx))


FLAG_REPEATED = const(1)


class MemoryReaderWriter:

    def __init__(self, buffer=None):
        self.buffer = buffer if buffer else []
        self.nread = 0
        self.nwritten = 0

    async def areadinto(self, buf):
        ln = len(buf)
        nread = min(ln, len(self.buffer))
        for idx in range(nread):
            buf[idx] = self.buffer.pop(0)
        self.nread += nread
        return nread

    async def awrite(self, buf):
        self.buffer.extend(buf)
        nwritten = len(buf)
        self.nwritten += nwritten
        return nwritten


class ElemRefObj:
    def __repr__(self):
        return 'RefObj'


class ElemRefArr:
    def __repr__(self):
        return 'RefAssoc'


def gen_elem_array(size, elem_type=None):
    """
    Generates element array of given size and initializes with given type.
    Supports container type, used for pre-allocation before deserialization.
    :param size:
    :param elem_type:
    :return:
    """
    if elem_type is None or not callable(elem_type):
        return [elem_type] * size
    if is_type(elem_type, ContainerType):
        elem_type = lambda: []
    res = []
    for _ in range(size):
        res.append(elem_type())
    return res


def is_elem_ref(elem_ref):
    """
    Returns true if the elem_ref is an element reference

    :param elem_ref:
    :return:
    """
    return elem_ref and isinstance(elem_ref, tuple) and len(elem_ref) == 3 \
           and (elem_ref[0] == ElemRefObj or elem_ref[0] == ElemRefArr)


def get_elem(elem_ref, default=None):
    """
    Gets the element referenced by elem_ref or returns the elem_ref directly if its not a reference.

    :param elem_ref:
    :param default:
    :return:
    """
    if not is_elem_ref(elem_ref):
        return elem_ref
    elif elem_ref[0] == ElemRefObj:
        return getattr(elem_ref[1], elem_ref[2], default)
    elif elem_ref[0] == ElemRefArr:
        return elem_ref[1][elem_ref[2]]


def set_elem(elem_ref, elem):
    """
    Sets element referenced by the elem_ref. Returns the elem.

    :param elem_ref:
    :param elem:
    :return:
    """
    if elem_ref is None or elem_ref == elem or not is_elem_ref(elem_ref):
        return elem

    elif elem_ref[0] == ElemRefObj:
        setattr(elem_ref[1], elem_ref[2], elem)
        return elem

    elif elem_ref[0] == ElemRefArr:
        elem_ref[1][elem_ref[2]] = elem
        return elem


def eref(obj, key, is_assoc=None):
    """
    Returns element reference
    :param obj:
    :param key:
    :param is_assoc:
    :return:
    """
    if obj is None:
        return None
    if isinstance(key, int) or (is_assoc is not None and is_assoc):
        return ElemRefArr, get_elem(obj), key
    else:
        return ElemRefObj, get_elem(obj), key


def get_ftype_params(field):
    """
    Convenient getter
    :param field:
    :return:
    """
    return field[1], field[2:]


def container_elem_type(container_type, params):
    """
    Returns container element type

    :param container_type:
    :param params:
    :return:
    """
    elem_type = params[0] if params else None
    if elem_type is None:
        elem_type = container_type.ELEM_TYPE
    return elem_type


class Archive(object):
    """
    Archive object for object binary serialization / deserialization.
    Resembles Archive API from the Monero codebase or Boost serialization archive.

    The design goal is to provide uniform API both for serialization and deserialization
    so the code is not duplicated for serialization and deserialization but the same
    for both ways in order to minimize potential bugs in the code.

    In order to use the archive for both ways we have to use so-called field references
    as we cannot directly modify given element as a parameter (value-passing) as its performed
    in C++ code. see: eref(), get_elem(), set_elem()
    """
    def __init__(self, iobj, writing=True, **kwargs):
        self.writing = writing
        self.iobj = iobj

    async def tag(self, tag):
        """

        :param tag:
        :return:
        """

    async def begin_array(self):
        """
        Mark start of the array. Used for JSON serialization.
        :return:
        """

    async def end_array(self):
        """
        Mark end of the array. Used for JSON serialization.
        :return:
        """

    async def begin_object(self):
        """
        Mark start of the object. Used for JSON serialization.
        :return:
        """

    async def end_object(self):
        """
        Mark end of the object. Used for JSON serialization.
        :return:
        """

    async def prepare_container(self, size, container, elem_type=None):
        """
        Prepares container for serialization
        :param size:
        :param container:
        :return:
        """
        if not self.writing:
            if container is None:
                return gen_elem_array(size, elem_type)

            fvalue = get_elem(container)
            if fvalue is None:
                fvalue = []
            fvalue += gen_elem_array(max(0, size - len(fvalue)), elem_type)
            set_elem(container, fvalue)
            return fvalue

    async def prepare_message(self, msg, msg_type):
        """
        Prepares message for serialization
        :param msg:
        :param msg_type:
        :return:
        """
        if self.writing:
            return
        return set_elem(msg, msg_type())

    async def uvarint(self, elem):
        """
        Uvarint
        :param elem:
        :return:
        """
        if self.writing:
            return await dump_uvarint(self.iobj, elem)
        else:
            return await load_uvarint(self.iobj)

    async def uint(self, elem, elem_type, params=None):
        """
        Fixed size int
        :param elem:
        :param elem_type:
        :param params:
        :return:
        """
        if self.writing:
            return await dump_uint(self.iobj, elem, elem_type.WIDTH)
        else:
            return await load_uint(self.iobj, elem_type.WIDTH)

    async def unicode_type(self, elem):
        """
        Unicode type
        :param elem:
        :return:
        """
        if self.writing:
            return await dump_unicode(self.iobj, elem)
        else:
            return await load_unicode(self.iobj)

    async def blob(self, elem=None, elem_type=None, params=None):
        """
        Loads/dumps blob
        :return:
        """
        elem_type = elem_type if elem_type else elem.__class__
        if hasattr(elem_type, 'serialize_archive'):
            elem = elem_type() if elem is None else elem
            return await elem.serialize_archive(self, elem=elem, elem_type=elem_type, params=params)

        if self.writing:
            return await dump_blob(self.iobj, elem=elem, elem_type=elem_type, params=params)
        else:
            return await load_blob(self.iobj, elem_type=elem_type, params=params, elem=elem)

    async def container(self, container=None, container_type=None, params=None):
        """
        Loads/dumps container
        :return:
        """
        if hasattr(container_type, 'serialize_archive'):
            container = container_type() if container is None else container
            return await container.serialize_archive(self, elem=container, elem_type=container_type, params=params)

        if self.writing:
            return await dump_container(self.iobj, container, container_type, params,
                                        field_archiver=self.dump_field)
        else:
            return await load_container(self.iobj, container_type, params=params, container=container,
                                        field_archiver=self.load_field)

    async def container_size(self, container_len=None, container_type=None, params=None):
        """
        Container size
        :param container_len:
        :param container_type:
        :param params:
        :return:
        """
        if hasattr(container_type, 'serialize_archive'):
            raise ValueError('not supported')

        if self.writing:
            return await dump_container_size(self.iobj, container_len, container_type, params)
        else:
            raise ValueError('Not supported')

    async def container_val(self, elem, container_type, params=None):
        """
        Single cont value
        :param elem:
        :param container_type:
        :param params:
        :param field_archiver:
        :return:
        """
        if hasattr(container_type, 'serialize_archive'):
            raise ValueError('not supported')
        if self.writing:
            return await dump_container_val(self.iobj, elem, container_type, params)
        else:
            raise ValueError('Not supported')

    async def tuple(self, elem=None, elem_type=None, params=None):
        """
        Loads/dumps tuple
        :return:
        """
        if hasattr(elem_type, 'serialize_archive'):
            container = elem_type() if elem is None else elem
            return await container.serialize_archive(self, elem=elem, elem_type=elem_type, params=params)

        if self.writing:
            return await dump_tuple(self.iobj, elem, elem_type, params,
                                    field_archiver=self.dump_field)
        else:
            return await load_tuple(self.iobj, elem_type, params=params, elem=elem,
                                    field_archiver=self.load_field)

    async def variant(self, elem=None, elem_type=None, params=None):
        """
        Loads/dumps variant type
        :param elem:
        :param elem_type:
        :param params:
        :return:
        """
        elem_type = elem_type if elem_type else elem.__class__
        if hasattr(elem_type, 'serialize_archive'):
            elem = elem_type() if elem is None else elem
            return await elem.serialize_archive(self, elem=elem, elem_type=elem_type, params=params)

        if self.writing:
            return await dump_variant(self.iobj, elem=elem,
                                      elem_type=elem_type if elem_type else elem.__class__,
                                      params=params, field_archiver=self.dump_field)
        else:
            return await load_variant(self.iobj, elem_type=elem_type if elem_type else elem.__class__,
                                      params=params, elem=elem, field_archiver=self.load_field)

    async def message(self, msg, msg_type=None):
        """
        Loads/dumps message
        :param msg:
        :param msg_type:
        :return:
        """
        elem_type = msg_type if msg_type is not None else msg.__class__
        if hasattr(elem_type, 'serialize_archive'):
            msg = elem_type() if msg is None else msg
            return await msg.serialize_archive(self)

        if self.writing:
            return await dump_message(self.iobj, msg, msg_type=msg_type, field_archiver=self.dump_field)
        else:
            return await load_message(self.iobj, msg_type, msg=msg, field_archiver=self.load_field)

    async def message_field(self, msg, field, fvalue=None):
        """
        Dumps/Loads message field
        :param msg:
        :param field:
        :param fvalue: explicit value for dump
        :return:
        """
        if self.writing:
            await dump_message_field(self.iobj, msg, field, fvalue=fvalue, field_archiver=self.dump_field)
        else:
            await load_message_field(self.iobj, msg, field, field_archiver=self.load_field)

    async def message_fields(self, msg, fields):
        """
        Load/dump individual message fields
        :param msg:
        :param fields:
        :param field_archiver:
        :return:
        """
        for field in fields:
            await self.message_field(msg, field)
        return msg

    async def rfield(self, elem=None, elem_type=None, params=None):
        """
        Loads/Dumps message field
        :param elem:
        :param elem_type:
        :param params:
        :return:
        """
        if self.writing:
            return await dump_field(self.iobj, elem=elem,
                                    elem_type=elem_type if elem_type else elem.__class__,
                                    params=params)
        else:
            return await load_field(self.iobj,
                                    elem_type=elem_type if elem_type else elem.__class__,
                                    params=params,
                                    elem=elem)

    async def field(self, elem=None, elem_type=None, params=None):
        """
        Archive field
        :param elem:
        :param elem_type:
        :param params:
        :return:
        """
        elem_type = elem_type if elem_type else elem.__class__
        fvalue = None
        if issubclass(elem_type, UVarintType):
            fvalue = await self.uvarint(get_elem(elem))

        elif issubclass(elem_type, IntType):
            fvalue = await self.uint(elem=get_elem(elem), elem_type=elem_type, params=params)

        elif issubclass(elem_type, BlobType):
            fvalue = await self.blob(elem=get_elem(elem), elem_type=elem_type, params=params)

        elif issubclass(elem_type, UnicodeType):
            fvalue = await self.unicode_type(get_elem(elem))

        elif issubclass(elem_type, VariantType):
            fvalue = await self.variant(elem=get_elem(elem), elem_type=elem_type, params=params)

        elif issubclass(elem_type, ContainerType):  # container ~ simple list
            fvalue = await self.container(container=get_elem(elem), container_type=elem_type, params=params)

        elif issubclass(elem_type, TupleType):  # tuple ~ simple list
            fvalue = await self.tuple(elem=get_elem(elem), elem_type=elem_type, params=params)

        elif issubclass(elem_type, MessageType):
            fvalue = await self.message(get_elem(elem), msg_type=elem_type)

        else:
            raise TypeError

        return fvalue if self.writing else set_elem(elem, fvalue)

    async def dump_field(self, writer, elem, elem_type, params=None):
        assert self.iobj == writer
        return await self.field(elem=elem, elem_type=elem_type, params=params)

    async def load_field(self, reader, elem_type, params=None, elem=None):
        assert self.iobj == reader
        return await self.field(elem=elem, elem_type=elem_type, params=params)


async def dump_blob(writer, elem, elem_type, params=None):
    """
    Dumps blob message to the writer.
    Supports both blob and raw value.

    :param writer:
    :param elem:
    :param elem_type:
    :param params:
    :return:
    """
    elem_is_blob = isinstance(elem, BlobType)
    elem_params = elem if elem_is_blob or elem_type is None else elem_type
    data = getattr(elem, BlobType.DATA_ATTR) if elem_is_blob else elem

    if not elem_params.FIX_SIZE:
        await dump_uvarint(writer, len(elem))
    elif len(data) != elem_params.SIZE:
        raise ValueError('Fixed size blob has not defined size: %s' % elem_params.SIZE)
    await writer.awrite(data)


async def load_blob(reader, elem_type, params=None, elem=None):
    """
    Loads blob from reader to the element. Returns the loaded blob.

    :param reader:
    :param elem_type:
    :param params:
    :param elem:
    :return:
    """
    ivalue = elem_type.SIZE if elem_type.FIX_SIZE else await load_uvarint(reader)
    fvalue = bytearray(ivalue)
    await reader.areadinto(fvalue)

    if elem is None:
        return fvalue  # array by default

    elif isinstance(elem, BlobType):
        setattr(elem, elem_type.DATA_ATTR, fvalue)
        return elem

    else:
        elem.extend(fvalue)

    return elem


async def dump_unicode(writer, elem):
    """
    Dumps string as UTF8 encoded string
    :param writer:
    :param elem:
    :return:
    """
    await dump_uvarint(writer, len(elem))
    await writer.awrite(bytes(elem, 'utf8'))


async def load_unicode(reader):
    """
    Loads UTF8 string
    :param reader:
    :return:
    """
    ivalue = await load_uvarint(reader)
    fvalue = bytearray(ivalue)
    await reader.areadinto(fvalue)
    return str(fvalue, 'utf8')


async def dump_container_size(writer, container_len, container_type, params=None):
    """
    Dumps container size - per element streaming
    :param writer:
    :param container:
    :param container_type:
    :param params:
    :param field_archiver:
    :return:
    """
    if not container_type or not container_type.FIX_SIZE:
        await dump_uvarint(writer, container_len)
    elif container_len != container_type.SIZE:
        raise ValueError('Fixed size container has not defined size: %s' % container_type.SIZE)


async def dump_container_val(writer, elem, container_type, params=None, field_archiver=None):
    """
    Single elem dump
    :param writer:
    :param elem:
    :param container_type:
    :param params:
    :return:
    """
    field_archiver = field_archiver if field_archiver else dump_field
    elem_type = container_elem_type(container_type, params)
    await field_archiver(writer, elem, elem_type, params[1:] if params else None)


async def dump_container(writer, container, container_type, params=None, field_archiver=None):
    """
    Dumps container of elements to the writer.

    :param writer:
    :param container:
    :param container_type:
    :param params:
    :param field_archiver:
    :return:
    """
    await dump_container_size(writer, len(container), container_type)

    field_archiver = field_archiver if field_archiver else dump_field
    elem_type = container_elem_type(container_type, params)

    for elem in container:
        await field_archiver(writer, elem, elem_type, params[1:] if params else None)


async def load_container(reader, container_type, params=None, container=None, field_archiver=None):
    """
    Loads container of elements from the reader. Supports the container ref.
    Returns loaded container.

    :param reader:
    :param container_type:
    :param params:
    :param container:
    :param field_archiver:
    :return:
    """
    field_archiver = field_archiver if field_archiver else load_field

    c_len = container_type.SIZE if container_type.FIX_SIZE else await load_uvarint(reader)
    if container and c_len != len(container):
        raise ValueError('Size mismatch')

    elem_type = container_elem_type(container_type, params)
    res = container if container else []
    for i in range(c_len):
        fvalue = await field_archiver(reader, elem_type,
                                      params[1:] if params else None,
                                      eref(res, i) if container else None)
        if not container:
            res.append(fvalue)
    return res


async def dump_tuple(writer, elem, elem_type, params=None, field_archiver=None):
    """
    Dumps tuple of elements to the writer.

    :param writer:
    :param elem:
    :param elem_type:
    :param params:
    :param field_archiver:
    :return:
    """
    if len(elem) != len(elem_type.FIELDS):
        raise ValueError('Fixed size tuple has not defined size: %s' % len(elem_type.FIELDS))
    await dump_uvarint(writer, len(elem))

    field_archiver = field_archiver if field_archiver else dump_field
    elem_fields = params[0] if params else None
    if elem_fields is None:
        elem_fields = elem_type.FIELDS
    for idx, elem in enumerate(elem):
        await field_archiver(writer, elem, elem_fields[idx], params[1:] if params else None)


async def load_tuple(reader, elem_type, params=None, elem=None, field_archiver=None):
    """
    Loads tuple of elements from the reader. Supports the tuple ref.
    Returns loaded tuple.

    :param reader:
    :param elem_type:
    :param params:
    :param container:
    :param field_archiver:
    :return:
    """
    field_archiver = field_archiver if field_archiver else load_field

    c_len = await load_uvarint(reader)
    if elem and c_len != len(elem):
        raise ValueError('Size mismatch')
    if c_len != len(elem_type.FIELDS):
        raise ValueError('Tuple size mismatch')

    elem_fields = params[0] if params else None
    if elem_fields is None:
        elem_fields = elem_type.FIELDS

    res = elem if elem else []
    for i in range(c_len):
        fvalue = await field_archiver(reader, elem_fields[i],
                                      params[1:] if params else None,
                                      eref(res, i) if elem else None)
        if not elem:
            res.append(fvalue)
    return res


async def dump_message_field(writer, msg, field, fvalue=None, field_archiver=None):
    """
    Dumps a message field to the writer. Field is defined by the message field specification.

    :param writer:
    :param msg:
    :param field:
    :param fvalue:
    :param field_archiver:
    :return:
    """
    fname, ftype, params = field[0], field[1], field[2:]
    fvalue = getattr(msg, fname, None) if fvalue is None else fvalue
    field_archiver = field_archiver if field_archiver else dump_field
    await field_archiver(writer, fvalue, ftype, params)


async def load_message_field(reader, msg, field, field_archiver=None):
    """
    Loads message field from the reader. Field is defined by the message field specification.
    Returns loaded value, supports field reference.

    :param reader:
    :param msg:
    :param field:
    :param field_archiver:
    :return:
    """
    fname, ftype, params = field[0], field[1], field[2:]
    field_archiver = field_archiver if field_archiver else load_field
    await field_archiver(reader, ftype, params, eref(msg, fname))


async def dump_message(writer, msg, msg_type=None, field_archiver=None):
    """
    Dumps message to the writer.

    :param writer:
    :param msg:
    :param msg_type:
    :param field_archiver:
    :return:
    """
    mtype = msg.__class__ if msg_type is None else msg_type
    fields = mtype.FIELDS
    if hasattr(mtype, 'serialize_archive'):
        raise ValueError('Cannot directly load, has to use archive with %s' % mtype)

    for field in fields:
        await dump_message_field(writer, msg=msg, field=field, field_archiver=field_archiver)


async def load_message(reader, msg_type, msg=None, field_archiver=None):
    """
    Loads message if the given type from the reader.
    Supports reading directly to existing message.

    :param reader:
    :param msg_type:
    :param msg:
    :param field_archiver:
    :return:
    """
    msg = msg_type() if msg is None else msg
    fields = msg_type.FIELDS if msg_type else msg.__class__.FIELDS
    if hasattr(msg_type, 'serialize_archive'):
        raise ValueError('Cannot directly load, has to use archive with %s' % msg_type)

    for field in fields:
        await load_message_field(reader, msg, field, field_archiver=field_archiver)

    return msg


async def dump_variant(writer, elem, elem_type=None, params=None, field_archiver=None):
    """
    Dumps variant type to the writer.
    Supports both wrapped and raw variant.

    :param writer:
    :param elem:
    :param elem_type:
    :param params:
    :param field_archiver:
    :return:
    """
    field_archiver = field_archiver if field_archiver else dump_field
    if isinstance(elem, VariantType) or elem_type.WRAPS_VALUE:
        await dump_uint(writer, elem.variant_elem_type.VARIANT_CODE, 1)
        await field_archiver(writer, getattr(elem, elem.variant_elem), elem.variant_elem_type)

    else:
        fdef = elem_type.find_fdef(elem_type.FIELDS, elem)
        await dump_uint(writer, fdef[1].VARIANT_CODE, 1)
        await field_archiver(writer, elem, fdef[1])


async def load_variant(reader, elem_type, params=None, elem=None, wrapped=None, field_archiver=None):
    """
    Loads variant type from the reader.
    Supports both wrapped and raw variant.

    :param reader:
    :param elem_type:
    :param params:
    :param elem:
    :param wrapped:
    :param field_archiver:
    :return:
    """
    is_wrapped = (isinstance(elem, VariantType) or elem_type.WRAPS_VALUE) if wrapped is None else wrapped
    if is_wrapped:
        elem = elem_type() if elem is None else elem

    field_archiver = field_archiver if field_archiver else load_field
    tag = await load_uint(reader, 1)
    for field in elem_type.FIELDS:
        ftype = field[1]
        if ftype.VARIANT_CODE == tag:
            fvalue = await field_archiver(reader, ftype, field[2:], elem if not is_wrapped else None)
            if is_wrapped:
                elem.set_variant(field[0], fvalue)
            return elem if is_wrapped else fvalue
    raise ValueError('Unknown tag: %s' % tag)


async def dump_field(writer, elem, elem_type, params=None):
    """
    Dumps field to the writer, according to the element specification.
    General multiplexer.

    :param writer:
    :param elem:
    :param elem_type:
    :param params:
    :return:
    """
    if issubclass(elem_type, UVarintType):
        await dump_uvarint(writer, elem)

    elif issubclass(elem_type, IntType):
        await dump_uint(writer, elem, elem_type.WIDTH)

    elif issubclass(elem_type, BlobType):
        await dump_blob(writer, elem, elem_type, params)

    elif issubclass(elem_type, UnicodeType):
        await dump_unicode(elem)

    elif issubclass(elem_type, VariantType):
        await dump_variant(writer, elem, elem_type, params)

    elif issubclass(elem_type, ContainerType):  # container ~ simple list
        await dump_container(writer, elem, elem_type, params)

    elif issubclass(elem_type, TupleType):  # container ~ simple list
        await dump_tuple(writer, elem, elem_type, params)

    elif issubclass(elem_type, MessageType):
        await dump_message(writer, elem, elem_type)

    else:
        raise TypeError


async def load_field(reader, elem_type, params=None, elem=None):
    """
    Loads a field from the reader, based on the field type specification. Demultiplexer.

    :param reader:
    :param elem_type:
    :param params:
    :param elem:
    :return:
    """
    if issubclass(elem_type, UVarintType):
        fvalue = await load_uvarint(reader)
        return set_elem(elem, fvalue)

    elif issubclass(elem_type, BoolType):
        fvalue = await load_uint(reader, elem_type.WIDTH)
        if fvalue != 0 and fvalue != 1:
            raise ValueError('Unexpected bool value')
        return set_elem(elem, fvalue)

    elif issubclass(elem_type, IntType):
        fvalue = await load_uint(reader, elem_type.WIDTH)
        return set_elem(elem, fvalue)

    elif issubclass(elem_type, BlobType):
        fvalue = await load_blob(reader, elem_type, params=params, elem=get_elem(elem))
        return set_elem(elem, fvalue)

    elif issubclass(elem_type, UnicodeType):
        fvalue = await load_unicode(reader)
        return set_elem(elem, fvalue)

    elif issubclass(elem_type, VariantType):
        fvalue = await load_variant(reader, elem_type, params=params, elem=get_elem(elem))
        return set_elem(elem, fvalue)

    elif issubclass(elem_type, ContainerType):  # container ~ simple list
        fvalue = await load_container(reader, elem_type, params=params, container=get_elem(elem))
        return set_elem(elem, fvalue)

    elif issubclass(elem_type, TupleType):  # tuple ~ simple list
        fvalue = await load_tuple(reader, elem_type, params=params, elem=get_elem(elem))
        return set_elem(elem, fvalue)

    elif issubclass(elem_type, MessageType):
        fvalue = await load_message(reader, msg_type=elem_type, msg=get_elem(elem))
        return set_elem(elem, fvalue)

    else:
        raise TypeError





#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Monero JSON codec.

Note: This is not optimized for memory usage.
Used for debugging.
'''

import base64
import collections
import json

from . import xmrserialize as x


class Archive(x.Archive):

    def __init__(self, iobj, writing=True, **kwargs):
        super().__init__(iobj, writing, **kwargs)


class AutoJSONEncoder(json.JSONEncoder):
    """
    JSON encoder trying to_json() first
    """
    def default(self, obj):
        try:
            return obj.to_json()
        except AttributeError:
            return self.default_classic(obj)

    def default_classic(self, o):
        if isinstance(o, set):
            return list(o)
        elif isinstance(o, bytes):
            return str(o, 'utf8')
        else:
            return super(AutoJSONEncoder, self).default(o)


def json_dumps(obj, **kwargs):
    """
    Uses auto encoder to serialize the object
    :param obj:
    :param kwargs:
    :return:
    """
    return json.dumps(obj, cls=AutoJSONEncoder, **kwargs)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
XMR RPC serialization
WARNING: Not finished yet

{json, binary} <-> portable_storage <-> object

The portable storage is pretty much generic dict in python.
The original codec is not streaming in a sense directly translating input to object or vice versa.

Portable storage is an intermediate object.

Mainly related to:

/contrib/epee/include/serialization/keyvalue_serialization.h
/contrib/epee/include/serialization/keyvalue_serialization_overloads.h
/contrib/epee/include/storages/portable_storage.h
/contrib/epee/include/storages/portable_storage_base.h
/contrib/epee/include/storages/portable_storage_from_bin.h
/contrib/epee/include/storages/portable_storage_from_json.h
/contrib/epee/include/storages/portable_storage_to_bin.h
/contrib/epee/include/storages/portable_storage_to_json.h

WARNING: Not finished yet
'''


from . import xmrserialize as x
from .xmrserialize import eref


_UINT_BUFFER = bytearray(1)


class PortableRawSizeMark:
    MASK = 0x03
    BYTE = 0
    WORD = 1
    DWORD = 2
    INT64 = 3


class SerializeType:
    INT64 = 1
    INT32 = 2
    INT16 = 3
    INT8 = 4
    UINT64 = 5
    UINT32 = 6
    UINT16 = 7
    UINT8 = 8
    DUOBLE = 9
    STRING = 10
    BOOL = 11
    OBJECT = 12
    ARRAY = 13
    ARRAY_FLAG = 0x80


SerializeTypeSize = {
    SerializeType.INT64: 8,
    SerializeType.INT32: 2,
    SerializeType.INT16: 2,
    SerializeType.INT8: 1,
    SerializeType.UINT64: 8,
    SerializeType.UINT32: 2,
    SerializeType.UINT16: 2,
    SerializeType.UINT8: 1,
    SerializeType.BOOL: 1,
}


XmrTypeMap = {
    x.UVarintType: SerializeType.UINT64,
    x.Int64: SerializeType.INT64,
    x.Int32: SerializeType.INT32,
    x.Int16: SerializeType.INT16,
    x.Int8: SerializeType.INT8,
    x.UInt64: SerializeType.UINT64,
    x.UInt32: SerializeType.UINT32,
    x.UInt16: SerializeType.UINT16,
    x.UInt8: SerializeType.UINT8,
    x.BoolType: SerializeType.BOOL,
    x.UnicodeType: SerializeType.STRING,
    x.ContainerType: SerializeType.ARRAY,
    x.MessageType: SerializeType.OBJECT,
}


def int_mark_to_size(int_type):
    if int_type == PortableRawSizeMark.BYTE:
        return 1
    elif int_type == PortableRawSizeMark.WORD:
        return 2
    elif int_type == PortableRawSizeMark.DWORD:
        return 4
    elif int_type == PortableRawSizeMark.INT64:
        return 8
    else:
        raise ValueError('Unknown int type')


def type_to_size(obj_type):
    return SerializeTypeSize[obj_type]


def xmr_type_to_type(elem_type):
    if isinstance(elem_type, int):
        return elem_type
    elif elem_type in XmrTypeMap:
        return XmrTypeMap[elem_type]
    elif isinstance(elem_type, x.XmrType):
        return XmrTypeMap[elem_type.__class__]
    else:
        raise ValueError('Cannot convert Xmr type')


async def dump_varint_t(writer, type_or, pv):
    """
    Binary dump of the integer of given type

    :param writer:
    :param type_or:
    :param pv:
    :return:
    """
    width = int_mark_to_size(type_or)
    n = (pv << 2) | type_or

    buffer = _UINT_BUFFER
    for _ in range(width):
        buffer[0] = n & 0xff
        await writer.awrite(buffer)
        n >>= 8

    return width


async def dump_varint(writer, val):
    """
    Binary dump of the variable size integer

    :param writer:
    :param val:
    :return:
    """
    if val <= 63:
        return dump_varint_t(writer, PortableRawSizeMark.BYTE, val)
    elif val <= 16383:
        return dump_varint_t(writer, PortableRawSizeMark.WORD, val)
    elif val <= 1073741823:
        return dump_varint_t(writer, PortableRawSizeMark.DWORD, val)
    else:
        if val > 4611686018427387903:
            raise ValueError('Int too big')
        return dump_varint_t(writer, PortableRawSizeMark.INT64, val)


async def load_varint(reader):
    """
    Binary load of variable size integer serialized by dump_varint

    :param reader:
    :return:
    """
    buffer = _UINT_BUFFER

    await reader.areadinto(buffer)
    width = int_mark_to_size(buffer[0] & PortableRawSizeMark.MASK)
    result = buffer[0]

    shift = 8
    for _ in range(width-1):
        await reader.areadinto(buffer)
        result += buffer[0] << shift
        shift += 8
    return result >> 2


async def dump_string(writer, val):
    """
    Binary string dump

    :param writer:
    :param val:
    :return:
    """
    await dump_varint(writer, len(val))
    await writer.awrite(bytes(val, 'utf8'))


async def load_string(reader):
    """
    Loads string from binary stream

    :param reader:
    :return:
    """
    ivalue = await load_varint(reader)
    fvalue = bytearray(ivalue)
    await reader.areadinto(fvalue)
    fvalue = str(fvalue, 'utf8')
    return fvalue


async def dump_blob(writer, elem, elem_type, params=None):
    """
    Dumps blob to a binary stream

    :param writer:
    :param elem:
    :param elem_type:
    :param params:
    :return:
    """
    elem_is_blob = isinstance(elem, x.BlobType)
    data = getattr(elem, x.BlobType.DATA_ATTR) if elem_is_blob else elem
    await dump_varint(writer, len(elem))
    await writer.awrite(data)


async def load_blob(reader, elem_type, params=None, elem=None):
    """
    Loads blob from reader to the element. Returns the loaded blob.

    :param reader:
    :param elem_type:
    :param params:
    :param elem:
    :return:
    """
    ivalue = await load_varint(reader)
    fvalue = bytearray(ivalue)
    await reader.areadinto(fvalue)

    if elem is None:
        return fvalue  # array by default

    elif isinstance(elem, x.BlobType):
        setattr(elem, elem_type.DATA_ATTR, fvalue)
        return elem

    else:
        elem.extend(fvalue)

    return elem


async def dump_container(writer, container, container_type, params=None, field_archiver=None):
    """
    Dumps container of elements to the writer.

    :param writer:
    :param container:
    :param container_type:
    :param params:
    :param field_archiver:
    :return:
    """
    field_archiver = field_archiver if field_archiver else dump_field
    elem_type = params[0] if params else None
    if elem_type is None:
        elem_type = container_type.ELEM_TYPE

    type_code = xmr_type_to_type(elem_type) | SerializeType.ARRAY_FLAG
    buff = bytearray(1)
    buff[0] = type_code
    await writer.awrite(buff)
    await dump_varint(writer, len(container))

    for elem in container:
        await field_archiver(writer, elem, elem_type, params[1:] if params else None)


#
# Code needs some check below this point
#


async def load_container(reader, container_type, params=None, container=None, field_archiver=None):
    """
    Loads container of elements from the reader. Supports the container ref.
    Returns loaded container.
    TODO: implement

    :param reader:
    :param container_type:
    :param params:
    :param container:
    :param field_archiver:
    :return:
    """
    field_archiver = field_archiver if field_archiver else load_field

    c_len = container_type.SIZE if container_type.FIX_SIZE else await load_varint(reader)
    if container and c_len != len(container):
        raise ValueError('Size mismatch')

    elem_type = params[0] if params else None
    if elem_type is None:
        elem_type = container_type.ELEM_TYPE

    res = container if container else []
    for i in range(c_len):
        fvalue = await field_archiver(reader, elem_type,
                                      params[1:] if params else None,
                                      eref(res, i) if container else None)
        if not container:
            res.append(fvalue)
    return res


async def dump_message_field(writer, msg, field, field_archiver=None):
    """
    Dumps a message field to the writer. Field is defined by the message field specification.
    TODO: implement

    :param writer:
    :param msg:
    :param field:
    :param field_archiver:
    :return:
    """
    fname, ftype, params = field[0], field[1], field[2:]
    fvalue = getattr(msg, fname, None)
    field_archiver = field_archiver if field_archiver else dump_field
    await field_archiver(writer, fvalue, ftype, params)


async def load_message_field(reader, msg, field, field_archiver=None):
    """
    Loads message field from the reader. Field is defined by the message field specification.
    Returns loaded value, supports field reference.
    TODO: implement

    :param reader:
    :param msg:
    :param field:
    :param field_archiver:
    :return:
    """
    fname, ftype, params = field[0], field[1], field[2:]
    field_archiver = field_archiver if field_archiver else load_field
    await field_archiver(reader, ftype, params, eref(msg, fname))


async def dump_message(writer, msg, field_archiver=None):
    """
    Dumps message to the writer.
    TODO: implement

    :param writer:
    :param msg:
    :param field_archiver:
    :return:
    """
    mtype = msg.__class__
    fields = mtype.FIELDS

    for field in fields:
        await dump_message_field(writer, msg=msg, field=field, field_archiver=field_archiver)


async def load_message(reader, msg_type, msg=None, field_archiver=None):
    """
    Loads message if the given type from the reader.
    Supports reading directly to existing message.
    TODO: implement

    :param reader:
    :param msg_type:
    :param msg:
    :param field_archiver:
    :return:
    """
    msg = msg_type() if msg is None else msg
    fields = msg_type.FIELDS if msg_type else msg.__class__.FIELDS
    for field in fields:
        await load_message_field(reader, msg, field, field_archiver=field_archiver)

    return msg


async def dump_field(writer, elem, elem_type, params=None):
    """
    TODO: implement

    :param writer:
    :param elem:
    :param elem_type:
    :param params:
    :return:
    """
    if issubclass(elem_type, x.UVarintType) or issubclass(elem_type, x.IntType) \
            or (elem_type is None and isinstance(elem, int)):
        await dump_varint(writer, elem)

    elif issubclass(elem_type, x.BlobType):
        await dump_blob(writer, elem, elem_type, params)

    elif issubclass(elem_type, x.UnicodeType):
        await dump_string(writer, elem)

    elif issubclass(elem_type, x.ContainerType):  # container ~ simple list
        await dump_container(writer, elem, elem_type, params)

    elif issubclass(elem_type, x.MessageType):
        await dump_message(writer, elem)

    else:
        raise TypeError


async def load_field(reader, elem_type, params=None, elem=None):
    """
    Loads a field from the reader, based on the field type specification. Demultiplexer.
    TODO: implement

    :param reader:
    :param elem_type:
    :param params:
    :param elem:
    :return:
    """
    if issubclass(elem_type, x.UVarintType) or issubclass(elem_type, x.IntType) \
            or (elem_type is None and isinstance(elem, int)):
        fvalue = await load_varint(reader)
        return x.set_elem(elem, fvalue)

    elif issubclass(elem_type, x.UnicodeType):
        fvalue = await load_string(reader)
        return x.set_elem(elem, fvalue)

    elif issubclass(elem_type, x.ContainerType):  # container ~ simple list
        fvalue = await load_container(reader, elem_type, params=params, container=x.get_elem(elem))
        return x.set_elem(elem, fvalue)

    elif issubclass(elem_type, x.MessageType):
        fvalue = await load_message(reader, msg_type=elem_type, msg=x.get_elem(elem))
        return x.set_elem(elem, fvalue)

    else:
        raise TypeError



